package ctbrec.recorder.download.dash;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.math.BigInteger;
import java.net.URL;
import java.nio.file.Path;
import java.text.DecimalFormat;
import java.time.Duration;
import java.time.Instant;
import java.util.ArrayList;
import java.util.List;
import java.util.Optional;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.TimeUnit;
import java.util.regex.Pattern;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBElement;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Unmarshaller;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import ctbrec.Config;
import ctbrec.Model;
import ctbrec.Recording;
import ctbrec.io.HttpClient;
import ctbrec.recorder.download.Download;
import ctbrec.recorder.download.dash.SegmentTimelineType.S;
import okhttp3.Request;
import okhttp3.Response;

public class DashDownload implements Download {
    private static final Logger LOG = LoggerFactory.getLogger(DashDownload.class);

    private int audioCounter = 0;
    private int videoCounter = 0;
    private DecimalFormat df = new DecimalFormat("00000");
    private boolean videoInitLoaded = false;
    private boolean audioInitLoaded = false;
    private BigInteger lastAudioTimestamp = BigInteger.ZERO;
    private BigInteger lastVideoTimestamp = BigInteger.ZERO;
    private Object downloadFinished = new Object();
    private HttpClient httpClient;
    private Config config;
    private Model model;
    private Instant startTime;
    private Instant endTime;
    private Path downloadDir;
    private String manifestUrl;
    private boolean running = false;

    private File targetFile;

    public DashDownload(HttpClient httpClient, String manifestUrl) {
        this.httpClient = httpClient;
        this.manifestUrl = manifestUrl;
    }

    private String getManifest(String url) throws IOException {
        // @formatter:off
        Request request = new Request.Builder()
                .url(url)
                .header("Accept", "*/*")
                .header("Accept-Language", "en-US,en;q=0.5")
                .header("User-Agent", config.getSettings().httpUserAgent)
                .header("Origin", model.getSite().getBaseUrl())
                .header("Referer", model.getSite().getBaseUrl())
                .header("Connection", "keep-alive")
                .build(); // @formatter:on
        LOG.trace("Loading manifest {}", url);
        try (Response response = httpClient.execute(request)) {
            return response.body().string();
        }
    }

    private int downloadSegments(MPDtype mpd, AdaptationSetType adaptationSet, boolean isVideo) throws IOException {
        if (adaptationSet == null) {
            return 0;
        }

        SegmentTemplateType segmentTemplate = adaptationSet.getSegmentTemplate();
        RepresentationType representation = chooseBestRepresentation(adaptationSet.getRepresentation());

        int downloaded = downloadInitChunksForVideoAndAudio(isVideo, mpd, segmentTemplate, representation);

        String media = segmentTemplate.getMedia();
        media = media.replaceAll("\\$RepresentationID\\$", representation.getId()); // NOSONAR

        List<S> segments = segmentTemplate.getSegmentTimeline().getS();
        if (!segments.isEmpty()) {
            BigInteger timestamp = segments.get(0).getT();
            BigInteger duration = BigInteger.ZERO;
            for (S s : segments) {
                timestamp = timestamp.add(duration);
                if (isVideo && timestamp.compareTo(lastVideoTimestamp) > 0 || !isVideo && timestamp.compareTo(lastAudioTimestamp) > 0) {
                    if (isVideo) {
                        lastVideoTimestamp = timestamp;
                    } else {
                        lastAudioTimestamp = timestamp;
                    }
                    String segmentUrl = media.replaceAll("\\$Time\\$", timestamp.toString());
                    duration = s.getD();
                    URL absUrl = new URL(new URL(mpd.getLocation().get(0)), segmentUrl);
                    download(downloadDir.toFile().getCanonicalPath(), absUrl, isVideo);
                    downloaded++;
                }
            }
        }

        return downloaded;
    }

    private RepresentationType chooseBestRepresentation(List<RepresentationType> representations) {
        RepresentationType best = null;
        long bestBandwidth = 0;
        for (RepresentationType rep : representations) {
            if (rep.getBandwidth() > bestBandwidth) {
                bestBandwidth = rep.getBandwidth();
                best = rep;
            }
        }
        return best;
    }

    private int downloadInitChunksForVideoAndAudio(boolean isVideo, MPDtype mpd, SegmentTemplateType segmentTemplate, RepresentationType representation)
            throws IOException {
        if (isVideo && !videoInitLoaded || !isVideo && !audioInitLoaded) {
            String initialization = segmentTemplate.getInitializationAttribute();
            initialization = initialization.replaceAll("\\$RepresentationID\\$", representation.getId());
            URL initUrl = new URL(new URL(mpd.getLocation().get(0)), initialization);
            download(downloadDir.toFile().getCanonicalPath(), initUrl, isVideo);
            if (isVideo) {
                LOG.debug("Video init chunk loaded");
                videoInitLoaded = true;
            } else {
                LOG.debug("Audio init chunk loaded");
                audioInitLoaded = true;
            }
            return 1;
        } else {
            return 0;
        }
    }

    private void download(String dir, URL url, boolean isVideo) throws IOException {
        // @formatter:off
        Request request = new Request.Builder()
                .url(url)
                .header("Accept", "*/*")
                .header("Accept-Language", "en-US,en;q=0.5")
                .header("User-Agent", config.getSettings().httpUserAgent)
                .header("Origin", model.getSite().getBaseUrl())
                .header("Referer", model.getSite().getBaseUrl())
                .header("Connection", "keep-alive")
                .build(); // @formatter:on
        int tries = 1;
        try (Response response = httpClient.execute(request)) {
            InputStream in = response.body().byteStream();
            String absFile = url.getFile();
            String prefix = isVideo ? "video" : "audio";
            int c = isVideo ? videoCounter++ : audioCounter++;
            File segmentFile = new File(dir, prefix + '_' + df.format(c) + '_' + new File(absFile).getName());
            while(tries <= 10) {
                if (!segmentFile.exists() || segmentFile.length() == 0) {
                    if (tries > 1) {
                        LOG.debug("Loading segment, try {}, {} {} {}", tries, response.code(), response.headers().values("Content-Length"), url);
                    } else {
                        LOG.trace("Loading segment, try {}, {} {} {}", tries, response.code(), response.headers().values("Content-Length"), url);
                    }
                    try (FileOutputStream out = new FileOutputStream(segmentFile)) {
                        byte[] b = new byte[1024];
                        int len = -1;
                        while ((len = in.read(b)) >= 0) {
                            out.write(b, 0, len);
                        }
                        out.flush();
                    }
                } else {
                    break;
                }
                tries++;
            }
        }
    }

    @Override
    public void init(Config config, Model model) {
        this.config = config;
        this.model = model;
        startTime = Instant.now();
        targetFile = Config.getInstance().getFileForRecording(model, "mp4");
        downloadDir = targetFile.getParentFile().toPath();
    }

    @Override
    public void start() throws IOException {
        try {
            running = true;
            JAXBContext jc = JAXBContext.newInstance(MPDtype.class.getPackage().getName());
            Unmarshaller u = jc.createUnmarshaller();
            while (running && !Thread.currentThread().isInterrupted()) {
                downloadManifestAndItsSegments(u);
            }
        } catch (Exception e) {
            LOG.error("Error while downloading dash stream", e);
        } finally {
            running = false;
            endTime = Instant.now();
            synchronized (downloadFinished) {
                downloadFinished.notifyAll();
            }
        }
    }

    private void downloadManifestAndItsSegments(Unmarshaller u) throws IOException, JAXBException, ExecutionException {
        String manifest = getManifest(manifestUrl);
        LOG.trace("Manifest: {}", manifest);
        @SuppressWarnings("unchecked")
        JAXBElement<MPDtype> root = (JAXBElement<MPDtype>) u.unmarshal(new ByteArrayInputStream(manifest.getBytes()));
        MPDtype mpd = root.getValue();
        List<PeriodType> periods = mpd.getPeriod();
        for (PeriodType period : periods) {
            List<AdaptationSetType> videoStreams = new ArrayList<>();
            List<AdaptationSetType> audioStreams = new ArrayList<>();

            List<AdaptationSetType> adaptationSets = period.getAdaptationSet();
            for (AdaptationSetType adaptationSet : adaptationSets) {
                String mimeType = adaptationSet.getMimeType();
                if (mimeType.equalsIgnoreCase("video/mp4")) {
                    videoStreams.add(adaptationSet);
                } else if (mimeType.equalsIgnoreCase("audio/mp4")) {
                    audioStreams.add(adaptationSet);
                }
            }

            downloadDir.toFile().mkdirs();

            AdaptationSetType video = chooseBestVideo(videoStreams);
            if (video == null) {
                throw new ExecutionException(new RuntimeException("No stream left in playlist"));
            } else {
                int downloaded = downloadSegments(mpd, video, true);

                AdaptationSetType audio = audioStreams.isEmpty() ? null : audioStreams.get(0);
                downloaded += downloadSegments(mpd, audio, false);

                if (downloaded == 0) {
                    LOG.trace("No new segments - Sleeping a bit");
                    waitSomeTime();
                }
            }
        }
    }

    private AdaptationSetType chooseBestVideo(List<AdaptationSetType> videoStreams) {
        AdaptationSetType best = null;
        int maxHeight = config.getSettings().maximumResolution;
        long bestHeight = 0;
        for (AdaptationSetType stream : videoStreams) {
            if (stream.getHeight() > bestHeight && (maxHeight == 0 || stream.getHeight() <= maxHeight)) {
                bestHeight = stream.getHeight();
                best = stream;
            }
        }
        return best;
    }

    @Override
    public void stop() {
        if (running) {
            internalStop();
            try {
                while(running) {
                    synchronized (downloadFinished) {
                        downloadFinished.wait(TimeUnit.SECONDS.toMillis(1));
                    }
                }
            } catch (InterruptedException e) {
                Thread.currentThread().interrupt();
                LOG.error("Couldn't wait for download to finish", e);
            }
        }
    }

    void internalStop() {
        running = false;
    }

    @Override
    public Model getModel() {
        return model;
    }

    @Override
    public Instant getStartTime() {
        return startTime;
    }

    @Override
    public Duration getLength() {
        return Duration.between(startTime, Optional.ofNullable(endTime).orElse(Instant.now()));
    }

    @Override
    public void postprocess(Recording recording) {
        try {
            new FfmpegMuxer(downloadDir.toFile(), targetFile.getName());
        } catch (IOException e) {
            LOG.error("Error while merging dash segments", e);
        }
    }

    @Override
    public File getTarget() {
        return targetFile;
    }

    @Override
    public String getPath(Model model) {
        String absolutePath = targetFile.getAbsolutePath();
        String recordingsDir = Config.getInstance().getSettings().recordingsDir;
        String relativePath = absolutePath.replaceFirst(Pattern.quote(recordingsDir), "");
        return relativePath;
    }

    /**
     * Causes the current thread to sleep for a short amount of time. This is used to slow down retries, if something is wrong with the playlist. E.g. HTTP 403
     * or 404
     */
    private void waitSomeTime() {
        try {
            Thread.sleep(1000);
        } catch (InterruptedException e) {
            Thread.currentThread().interrupt();
        }
    }

}
